import os
import sys
import argparse
import h5py
import numpy
import numpy as np
import matplotlib as mpl
mpl.use("Agg")
import matplotlib.pyplot as plt
import corner

labels = {
    "sigma_cos" : r"$\sigma$",
    "E_chi" : r"$\mathbb{E}[\chi]$",
    "Var_chi" : r"$\mathrm{Var}[\chi]$",
}
ranges = {
    "sigma_cos" : (0.0, 4.0),
    "E_chi" : (0.0, 1.0),
    "Var_chi" : (0.0, 0.25),
}

param_names = ["E_chi", "Var_chi", "sigma_cos"]

suffices = ["1_pl0", "2_pl0", "1_g0", "2_g0", "prior"]
colors = ["C0", "C1", "C2", "C3", "black"]
linestyles = ["solid", "solid", "solid", "solid", "dashed"]

n_samples_prior = 1 << 24

random_state = numpy.random.RandomState()

def make_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("posteriors")
    parser.add_argument("output")

    parser.add_argument("--seed", type=int)

    return parser


def main():
    cli_parser = make_parser()
    cli_args = cli_parser.parse_args()

    random_state = numpy.random.RandomState(cli_args.seed)

    with h5py.File(cli_args.posteriors, "r") as posteriors_file:
        posteriors = {
            "E_chi" : load_params(posteriors_file, "E_chi"),
            "Var_chi" : load_params(posteriors_file, "Var_chi"),
            "sigma_cos" : load_params(posteriors_file, "sigma_cos"),
        }

    # Add in priors
    E_chi_prior, Var_chi_prior = (
        draw_prior_E_Var_chi(n_samples_prior, random_state=random_state)
    )
    sigma_cos_prior = (
        draw_prior_sigma_cos(n_samples_prior, random_state=random_state)
    )
    posteriors["E_chi"]["prior"] = E_chi_prior
    posteriors["Var_chi"]["prior"] = Var_chi_prior
    posteriors["sigma_cos"]["prior"] = sigma_cos_prior

    fig, axes = plt.subplots(
        1, 3,
        constrained_layout=True,
        figsize=(8, 3),
    )
    axes_dict = dict(zip(param_names, axes))
    axes[0].set_ylabel("PDF")

    for param_name in param_names:
        ax = axes_dict[param_name]
        rnge = ranges[param_name]
        label = labels[param_name]

        ax.set_xlim(*rnge)
        ax.set_yticklabels([])

        for color, linestyle, suffix in zip(colors, linestyles, suffices):
            samples = posteriors[param_name][suffix]

            ax.hist(
                samples, bins=20, range=rnge,
                color=color, linestyle=linestyle,
                histtype="step", density=True,
            )

        ax.set_xlabel(label)

    fig.savefig(cli_args.output)


def load_params(posteriors_file, prefix):
    param_names = posteriors_file["param_names"]
    variable_names = [
        name.decode() for name in param_names
        if (name not in posteriors_file["constants"].attrs) and
           (name not in posteriors_file["duplicates"].attrs)
    ]
    samples = posteriors_file["samples"][()].T

    variables = dict(zip(variable_names, samples))

    return {
        name[len(prefix):] : values
        for name, values in variables.items()
        if name.startswith(prefix)
    }


def draw_prior_E_Var_chi(n_samples, random_state=random_state):
    E_chi_min, E_chi_max = ranges["E_chi"]
    Var_chi_min, Var_chi_max = ranges["Var_chi"]

    E_chi = numpy.empty(n_samples)
    Var_chi = numpy.empty(n_samples)

    i_start = 0
    n_bad = n_samples

    while n_bad > 0:
        E_chi_cand = random_state.uniform(E_chi_min, E_chi_max, n_samples)
        Var_chi_cand = random_state.uniform(Var_chi_min, Var_chi_max, n_samples)

        alpha_chi_cand, beta_chi_cand = (
            mean_variance_to_alpha_beta(E_chi_cand, Var_chi_cand)
        )

        i_good_cand = (alpha_chi_cand >= 1.0) & (beta_chi_cand >= 1.0)
        n_good_cand = numpy.count_nonzero(i_good_cand)

        alpha_chi_cand_good = alpha_chi_cand[i_good_cand]
        beta_chi_cand_good = beta_chi_cand[i_good_cand]

        if n_good_cand > n_bad:
            alpha_chi_cand_good = alpha_chi_cand_good[:n_bad]
            beta_chi_cand_good = beta_chi_cand_good[:n_bad]
            n_good_cand = n_bad

        E_chi_cand_good, Var_chi_cand_good = (
            alpha_beta_to_mean_variance(alpha_chi_cand_good, beta_chi_cand_good)
        )

        E_chi[i_start:i_start+n_good_cand] = E_chi_cand_good
        Var_chi[i_start:i_start+n_good_cand] = Var_chi_cand_good

        i_start += n_good_cand
        n_bad -= n_good_cand

    return E_chi, Var_chi


def draw_prior_sigma_cos(n_samples, random_state=random_state):
    sigma_min, sigma_max = ranges["sigma_cos"]
    return random_state.uniform(sigma_min, sigma_max, n_samples)


def mean_variance_to_alpha_beta(
        mean, variance, loc=0.0, scale=1.0,
        xpy=numpy,
    ):
    import numpy

    a, c = loc, loc+scale

    a_minus_mean = a - mean
    c_minus_mean = c - mean

    shared_term = (
        (a_minus_mean*c_minus_mean + variance) /
        (scale * variance)
    )

    alpha =  shared_term * a_minus_mean
    beta  = -shared_term * c_minus_mean

    return alpha, beta


def alpha_beta_to_mean_variance(
        alpha, beta, loc=0.0, scale=1.0,
        xpy=numpy,
    ):
    import numpy

    a, c = loc, loc+scale
    alpha_plus_beta = alpha + beta

    mean = (alpha*c + beta*a) / alpha_plus_beta

    variance = (
        xpy.square(scale) * alpha*beta /
        (xpy.square(alpha_plus_beta) * (alpha_plus_beta + 1.0))
    )

    return mean, variance


if __name__ == "__main__":
    sys.exit(main())
